from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


class ModuleDocFragment(object):
    DOCUMENTATION = r'''
options:
  normal_doc_frag:
    description:
    - an option
'''

    OTHER_DOCUMENTATION = r'''
options:
  other_doc_frag:
    description:
    - another option
'''
